#!/usr/bin/python
# $Id: steam.py,v 1.9 2006/03/31 12:39:05 alex Exp $

import math
import random
import pygame
import sys
from OpenGL.GL import *
from OpenGL.GLUT import *

import res
import collide

class Sprite:
    def __init__(self, filename, pos=(0,0)):
        self.tex, self.width, self.height, self.u, self.v \
            = res.getTexture(filename)
        self.hx, self.hy = self.width/2, self.height/2
        self.pos = pos

    def draw(self, x=None, y=None):
        glBindTexture(GL_TEXTURE_2D, self.tex)
        if x == None or y == None:
            self.drawQuad(self.pos[0], self.pos[1])
        else:
            self.drawQuad(x, y)

    def drawQuad(self, x, y):
        glBegin(GL_QUADS)
        glTexCoord(0, 0)
        glVertex(x-self.hx, y-self.hy)
        glTexCoord(0, self.v)
        glVertex(x-self.hx, y+self.hy)
        glTexCoord(self.u, self.v)
        glVertex(x+self.hx, y+self.hy)
        glTexCoord(self.u, 0)
        glVertex(x+self.hx, y-self.hy)
        glEnd()

class MetalBehaviour(collide.Behaviour):
    def __init__(self, dropGrid, steamGrid):
        self.grids = [dropGrid, steamGrid]
    
    def respondDrop(self, collider, point):
        return self.DROP_BOUNCE

    def respondSteam(self, collider, point):
        if random.random() < 0.05:
            # TODO
            game.droplets.addDroplet(point[0], point[1], collider)
            return self.STEAM_CONDENSE
        else:
            return self.STEAM_BOUNCE

class WoodBehaviour(collide.Behaviour):
    def __init__(self, dropGrid, steamGrid):
        self.grids = [dropGrid, steamGrid]

    def respondDrop(self, collider, point):
        return self.DROP_REMOVE

    def respondSteam(self, collider, point):
        return self.STEAM_BOUNCE

class VaporiseBehaviour(collide.Behaviour):
    def __init__(self, dropGrid, steamGrid):
        self.grids = [dropGrid, steamGrid]

    def respondDrop(self, collider, point):
        game.particles.addParticle(point)
        return self.DROP_REMOVE

    def respondSteam(self, collider, point):
        return self.STEAM_BOUNCE

class PlaceableObject(collide.PolygonCollider):
    DRAG_NULL = 0
    DRAG_TRANSLATE = 1
    DRAG_ROTATE = 2

    def __init__(self, width, height, texture, behaviour=None):
        collide.PolygonCollider.__init__(self, 
                                         [(0, 0), (-1, -1), (0, -1)], 
                                         behaviour)
        self.width, self.height = width, height
        # Start off-screen
        self.x, self.y = -width, -height
        self.rotation = 0
        self.rotationHandleSprite = Sprite('rot_handle.png')
        self.dragOp = self.DRAG_NULL

        self.selectable = True
        self.hoverable = True

    def setPosition(self, x, y):
        self.x, self.y = x, y
        self.updateCollisionPolygon()

    def setRotation(self, rotation):
        self.rotation = rotation
        self.updateCollisionPolygon()

    def updateCollisionPolygon(self):
        # Yes this is very verbose and could be done in a couple of lines
        # but it would be really embarassing to make a mistake here!
        hx, hy = self.width/2, self.height/2
        rads = self.rotation*math.pi/180
        points = [
            [-hx, -hy], 
            [hx,  -hy],
            [hx,   hy],
            [-hx,  hy]]
        rotMatrix = (
            (math.cos(rads), -math.sin(rads)),
            (math.sin(rads), math.cos(rads)))
        for point in points:
            x, y = point
            point[0] = x*rotMatrix[0][0] + y*rotMatrix[0][1]
            point[1] = x*rotMatrix[1][0] + y*rotMatrix[1][1]
        for point in points:
            point[0] += self.x
            point[1] += self.y
        self.setPoints(points)

    def draw(self):
        glDisable(GL_TEXTURE_2D)
        glColor(0.8, 1.0, 0.9, 1.0)
        glBegin(GL_POLYGON)
        for point in self.points:
            glVertex(point[0], point[1], 0)
        glEnd()
        glEnable(GL_TEXTURE_2D)

    def drawSelection(self):
        radius = max(self.width, self.height)/2 + 10

        glPushMatrix()
        glTranslate(self.x, self.y, 0)
        """
        # Draw faint circle around object
        circlePoints = 20
        glDisable(GL_TEXTURE_2D)
        glColor(0.7,0.7,0.7,0.8)
        glBegin(GL_LINE_LOOP)
        for i in range(circlePoints):
            angle = float(i) / circlePoints * 2 * math.pi
            glVertex(math.cos(angle) * radius, math.sin(angle) * radius)
        glEnd()
        glEnable(GL_TEXTURE_2D)
        """

        # Draw rotation handles
        glRotate(self.rotation, 0, 0, 1)
        for i in range(4):
            angle = float(i) / 4 * 2 * math.pi + math.pi/4
            glPushMatrix()
            glTranslate(math.cos(angle) * radius, math.sin(angle) * radius, 0)
            glRotate(float(i)/4 * 360 - 45, 0, 0, 1)
            self.rotationHandleSprite.draw()
            glPopMatrix()
        glPopMatrix()

    def keepHover(self, x, y):
        """ There is some tolerance when mouse moves out of selection before
            dropping it """
        radius2 = (max(self.width, self.height)/2 + 15)**2
        return (x-self.x)**2 + (y-self.y)**2 < radius2

    def beginDrag(self, x, y, alternateButton):
        if self.containsPoint(x, y):
            self.dragOp = self.DRAG_TRANSLATE
            self.dragOffset = x - self.x, y - self.y
        else:
            self.dragOp = self.DRAG_ROTATE
            self.dragOffsetAngle = math.atan2(y-self.y, x-self.x)*180/math.pi \
                                    - self.rotation

    def continueDrag(self, x, y):
        if self.dragOp == self.DRAG_TRANSLATE:
            self.x, self.y = x - self.dragOffset[0], y - self.dragOffset[1]
        elif self.dragOp == self.DRAG_ROTATE:
            self.rotation = math.atan2(y-self.y, x-self.x)*180/math.pi \
                                - self.dragOffsetAngle
        self.updateCollisionPolygon()

    def completeDrag(self, x, y):
        self.dragOp = self.DRAG_NULL
        self.updateGrids()

class LineStrip(collide.LineStripCollider):
    POINT_RENDER_SIZE = 2

    def __init__(self, points, behaviour):
        collide.LineStripCollider.__init__(self, points, behaviour)
        self.hoverable = True
        self.selectable = True

    def containsPoint(self, x, y):
        if self.getNodePointAt(x, y) is not None:
            return True
        return False
    
    def getNodePointAt(self, x, y):
        sz = self.POINT_RENDER_SIZE
        for point in self.points:
            if x >= point[0]-sz and x <= point[0]+sz and \
               y >= point[1]-sz and y <= point[1]+sz:
                return point
        return None

    def draw(self, selected=False):
        glDisable(GL_TEXTURE_2D)
        glColor(0, 0, 1)
        glBegin(GL_LINE_STRIP)
        for point in self.points:
            glVertex(point[0], point[1], 0)
        glEnd()
        sz = self.POINT_RENDER_SIZE
        glColor(1, 0, 0)
        if selected:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        else:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glBegin(GL_QUADS)
        for point in self.points:
            glVertex(point[0]-sz, point[1]-sz)
            glVertex(point[0]+sz, point[1]-sz)
            glVertex(point[0]+sz, point[1]+sz)
            glVertex(point[0]-sz, point[1]+sz)
        glEnd()
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glEnable(GL_TEXTURE_2D)

    def drawSelection(self):
        self.draw(True)

    def keepHover(self, x, y):
        return self.containsPoint(x, y)

    def beginDrag(self, x, y, alternateButton):
        if alternateButton:
            # Create new node
            self.dragPoint = [x, y]
            if self.getNodePointAt(x, y) == self.points[0]:
                # At start
                otherPoint = self.points[0]
                self.points.insert(0, self.dragPoint)
            else:
                # At end
                otherPoint = self.points[-1]
                self.points.append(self.dragPoint)
            self.colliders.append(collide.SegmentCollider(self.dragPoint, 
                                                          otherPoint, 
                                                          self.behaviour))
        else:
            self.dragPoint = self.getNodePointAt(x, y)
        self.dragOffset = x - self.dragPoint[0], y - self.dragPoint[1]

    def continueDrag(self, x, y):
        self.dragPoint[0] = x - self.dragOffset[0]
        self.dragPoint[1] = y - self.dragOffset[1]

    def completeDrag(self, x, y):
        self.updateGrids()
        self.interactiveGrid.updateCollider(self, self.getBoundingBox())

    def addInteractiveToGrid(self, grid):
        self.interactiveGrid = grid
        grid.addCollider(self, self.getBoundingBox())

    def getBoundingBox(self):
        # Only used for interactive (realtime is segments)
        x1, y1, x2, y2 = (self.points[0][0], self.points[0][1],
                          self.points[0][0], self.points[0][1])
        for point in self.points:
            x1 = min(x1, point[0])
            y1 = min(y1, point[1])
            x2 = max(x2, point[0])
            y2 = max(y2, point[1])
        return x1, y1, x2, y2

class MetalObject(PlaceableObject):
    def __init__(self, x, y, behaviour):
        PlaceableObject.__init__(self, 90, 20, 'metal.png', behaviour)
        self.setPosition(x, y)

class WoodObject(PlaceableObject):
    def __init__(self, x, y, behaviour):
        PlaceableObject.__init__(self, 90, 20, 'wood.png', behaviour)
        self.setPosition(x, y)

class PlantBehaviour(collide.Behaviour):
    def __init__(self, plant, dropGrid):
        self.plant = plant
        self.grids = [dropGrid]

    def respondDrop(self, collider, point):
        self.plant.dropCount += 1
        return self.DROP_REMOVE

class Plant(PlaceableObject):
    potAttachment = (0,13)
    def __init__(self, x, y, dropGrid):
        self.potSprite = Sprite('pot2.png')
        PlaceableObject.__init__(self, self.potSprite.width, 20, '')
        self.behaviour = PlantBehaviour(self, dropGrid)
        self.setPosition(x, y)
        self.plantSegmentSprite = Sprite('plant_seg1.png')
        self.plantFlowerSprite = Sprite('flower.png')
        self.droop = 30 
        self.dropCountDeriv = 0.0
        self.dropCount = 0 
        self.segments = 5

    def draw(self):
        glPushMatrix()
        glTranslate(self.x, self.y, 0)
        glColor(1, 1, 1, 1)     # XXX or turn off modulate for this tex?
        self.potSprite.draw()
        glBindTexture(GL_TEXTURE_2D, self.plantSegmentSprite.tex)
        glTranslate(self.potSprite.pos[0] + self.potAttachment[0],
                    self.potSprite.pos[1] + self.potAttachment[1], 0)
        advance = self.plantSegmentSprite.hy-2
        for i in range(self.segments):
            glRotate(self.droop, 0, 0, 1)
            glTranslate(0, advance, 0)
            self.plantSegmentSprite.drawQuad(0, 0)
            glTranslate(0, advance, 0)
        self.plantFlowerSprite.draw()
        glPopMatrix()

    def update(self, millis):
        if millis == 0:
            return
        # Some kind of derivative fall-off function which jumped into my head
        # and I can't really explain now.
        self.dropCountDeriv = self.dropCountDeriv * 0.8 \
                                    + self.dropCount / float(millis) * 0.1

        # LERP from 0 to 30 degrees for 5 drips/sec to 0 drips/sec
        targetDroop = max(min(30 - self.dropCountDeriv*1000/5*30, 30), 0)

        # Constant velocity to target droop
        e = 0.5
        if self.droop < targetDroop - e:
            self.droop += 0.003 * millis
        elif self.droop > targetDroop + e:
            self.droop -= 0.003 * millis
        self.dropCount = 0


        
class Droplets:
    X=0
    Y=1
    DX=2
    DY=3
    COLLIDER=4
    COMMITTED=5
    DEBUG=6
    def __init__(self, filename, acceleration, 
                 tangentAcceleration, drag, count):
        self.tex, self.width, self.height, self.u, self.v \
            = res.getTexture(filename)
        self.acceleration = acceleration
        self.tangentAcceleration = tangentAcceleration
        self.drag = drag
        self.count = count
        self.droplets = []
        self.radius = max(self.width/2, self.height/2)

    def addDroplet(self, x, y, collider):
        if len(self.droplets) < self.count:
            self.droplets.append([x, y, 0, 0, collider, False, None])

    def draw(self):
        glBindTexture(GL_TEXTURE_2D, self.tex)
        for x, y, dx, dy, collider, committed, debug in self.droplets:
            glPushMatrix()
            glColor(1, 1, 1, 1)
            glTranslate(x, y, 0)
            glEnable(GL_TEXTURE_2D)
            glBegin(GL_QUADS)
            glTexCoord(0, 0)
            glVertex(-self.radius, -self.radius)
            glTexCoord(0, self.v)
            glVertex(-self.radius, self.radius)
            glTexCoord(self.u, self.v)
            glVertex(self.radius, self.radius)
            glTexCoord(self.u, 0)
            glVertex(self.radius, -self.radius)
            glEnd()

            if debug:
                dx, dy = debug[0] - x, debug[1] - y
                glDisable(GL_TEXTURE_2D)
                glBegin(GL_LINES)
                glColor(1, 0, 0, 1)
                glVertex(0, 0, 0)
                glColor(0, 1, 0, 1)
                glVertex(dx, dy, 0)
                glEnd()
            
            glPopMatrix()
        
    def update(self, millis, grid):
        for drop in self.droplets:
            collided = False
            # If the drop is sliding along a collider, do collision with that
            # collider first to make sure that
            #  a) it's still nearby (haven't fallen off the edge)
            #  b) the drop velocity is tangent to the collider's edge
            if drop[self.COLLIDER]:
                collision = drop[self.COLLIDER].collideSphere(drop[self.X], 
                                                   drop[self.Y], 
                                                   self.radius)
                if collision:
                    penetration, tangent, nearest, collider = collision
                    tx, ty = tangent
                    # Find direction of tangent closest to acceleration
                    dot = tx*self.acceleration[0] + \
                          ty*self.acceleration[1]
                    if dot < 0:
                        tx, ty = -tx, -ty
                    # Normalize tangent
                    mag = math.sqrt(tx**2 + ty**2)
                    tx, ty = tx/mag, ty/mag
                    if ty == 0:
                        # Tangent is horizontal, just drop the particle
                        drop[self.DX] = 0
                        drop[self.DY] = 0
                        drop[self.COLLIDER] = None
                    else:
                        # Velocity along tangent. randomness prevents clumping
                        a = random.random() * 0.5 + 0.5  
                        ddy = self.tangentAcceleration * ty * millis * a
                        ddx = self.tangentAcceleration * tx * millis * a
                        dx = drop[self.DX] + ddx
                        dy = drop[self.DY] + ddy
                        drop[self.DX] = dx
                        drop[self.DY] = dy

                        # Ray-test the new trajectory with other colliders and
                        # possibly transfer the drop to another collider.
                        # Don't do this if we just switched to a new collider
                        # (committed)
                        if not drop[self.COMMITTED]:
                            x2 = drop[self.X] + dx * millis
                            y2 = drop[self.Y] + dy * millis
                            # TODO ignore drop's collider early
                            collisions = grid.collideRay(
                                            drop[self.X], drop[self.Y], x2, y2)
                            for collision in collisions:
                                penetration, tangent, nearest, collider \
                                    = collision
                                if collider is not drop[self.COLLIDER]:
                                    # Have slid along to the intersection with
                                    # another collider.  Switch over to the
                                    # new collider and stop so the next frame
                                    # will have the correct velocity.
                                    drop[self.COLLIDER] = collider
                                    drop[self.COMMITTED] = True
                                    drop[self.X], drop[self.Y] = nearest
                                    drop[self.DX] = 0
                                    drop[self.DY] = 0
                                    break
                        else:
                            drop[self.COMMITTED] = False
                else:
                    # No longer colliding, make it free-fall
                    drop[self.COLLIDER] = None

            if not drop[self.COLLIDER]:
                # Free-falling
                # Check for intersection in this time-step by approximating
                # trajectory to a ray
                dx = drop[self.DX] + self.acceleration[0] * millis
                dy = drop[self.DY] + self.acceleration[1] * millis
                dx -= dx * self.drag[0] * millis
                dy -= dy * self.drag[1] * millis
                x2 = drop[self.X] + dx * millis
                y2 = drop[self.Y] + dy * millis
                collisions = grid.collideRay(drop[self.X], drop[self.Y],
                                                 x2, y2)
                collision = grid.closestCollision(collisions)
                if collision and dx**2 + dy**2 > 0:
                    # Found ray collision.. bounce or slide or vaporize
                    penetration, tangent, nearest, collider = collision
                    response = collider.respondDrop(collider, nearest)
                    if response == collide.Behaviour.DROP_REMOVE:
                        # Filter at end of loop will deactivate it
                        drop[self.Y] = -20  
                    elif response == collide.Behaviour.DROP_BOUNCE:
                        # Make sure normal is opposite dir to velocity
                        nx, ny = -tangent[1], tangent[0]
                        mag = math.sqrt(nx**2 + ny**2)
                        nx, ny = nx/mag, ny/mag
                        dot = nx*dx + ny*dy
                        if dot > 0:
                            nx, ny = -nx, -ny
                            dot *= -1
                        rx = dx - 2 * nx * dot
                        ry = dy - 2 * ny * dot
                        drop[self.DX] = rx * 0.9
                        drop[self.DY] = ry * 0.9
                        # A bit dodgy.. warp droplet to collision point
                        drop[self.X], drop[self.Y] = nearest
                    else:
                        assert False
                        
                drop[self.DX] += self.acceleration[0] * millis
                drop[self.DY] += self.acceleration[1] * millis
                drop[self.DX] -= drop[self.DX] * self.drag[0] * millis
                drop[self.DY] -= drop[self.DY] * self.drag[1] * millis
            drop[self.X] += drop[self.DX] * millis
            drop[self.Y] += drop[self.DY] * millis
        # TODO is this best way to remove/deactivate dead drops?
        self.droplets = [drop for drop in self.droplets if drop[self.Y] > 0]


class Particles:
    # Each particle is array indexed with:
    X=0
    Y=1
    DX=2
    DY=3
    ACTIVE=4
    TTL=5
    SCALE=6
    CONDENSED=7
    
    def __init__(self, filename, 
                 count, rate, lifetime, origin, velocity, velocitySigma,
                 scaleVelocity, buoyancy, buoyancyAcceleration):
        """ rate: millis between particle creation """
        self.origin = origin
        self.velocity = velocity
        self.velocitySigma = velocitySigma
        self.buoyancyAcceleration = buoyancyAcceleration
        self.buoyancy = buoyancy
        self.scaleVelocity = scaleVelocity
        self.rate = rate
        self.lifetime = lifetime
        self.particles = [[0, 0, 0, 0, False, 0, 1.0, False] for i in range(count)]
        self.tex, self.width, self.height, self.u, self.v \
            = res.getTexture(filename)
        self.particleWaitTime = 0
        self.radius=max(self.width/2, self.height/2)
        self.originQueue = []

    def addParticle(self, point):
        self.originQueue.append(point)

    def draw(self):
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        glBindTexture(GL_TEXTURE_2D, self.tex)
        hx = self.width/2
        hy = self.height/2
        for x, y, dx, dy, active, ttl, scale, condensed in self.particles:
            if not active:
                continue
            # TODO can avoid push,pop, translate and scale and do everything in
            # one big begin/end by moving the maths into python
            glPushMatrix()
            glColor(1, 1, 1, max(scale,1.0) * float(ttl)/self.lifetime)
            glTranslate(x, y, 0)
            glScale(scale, scale, 1)
            glBegin(GL_QUADS)
            glTexCoord(0, 0)
            glVertex(-hx, -hy)
            glTexCoord(0, self.v)
            glVertex(-hx, hy)
            glTexCoord(self.u, self.v)
            glVertex(hx, hy)
            glTexCoord(self.u, 0)
            glVertex(hx, -hy)
            glEnd()  
            glPopMatrix()
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)

    def update(self, millis, grid):
        self.particleWaitTime += millis
        particlesToAdd = self.particleWaitTime / self.rate
        self.particleWaitTime -= self.rate * particlesToAdd
        for particle in self.particles:
            if not particle[self.ACTIVE] and particlesToAdd > 0:
                if self.originQueue:
                    origin = self.originQueue.pop()
                else:
                    # TODO round-robin on origins
                    origin = self.origin
                particle[self.X] = origin[0]
                particle[self.Y] = origin[1]
                # TODO: performance of gauss vs uniform?
                particle[self.DX] = \
                    random.gauss(self.velocity[0], self.velocitySigma[0])
                particle[self.DY] = \
                    random.gauss(self.velocity[1], self.velocitySigma[1])
                particle[self.ACTIVE] = True
                particle[self.TTL] = self.lifetime
                particle[self.SCALE] = 0.0
                particle[self.CONDENSED] = False
                particlesToAdd -= 1
            elif not particle[self.ACTIVE]:
                continue

            # This test is a roundabout way of discouraging particles that
            # are bouncing against a wall from being buoyant.
            if particle[self.SCALE] > 0.5:
                u = millis*self.buoyancyAcceleration
                particle[self.DX] = particle[self.DX]*(1-u) + u*self.buoyancy[0]
                particle[self.DY] = particle[self.DY]*(1-u) + u*self.buoyancy[1]
            particle[self.X] += particle[self.DX] * millis
            particle[self.Y] += particle[self.DY] * millis
            particle[self.TTL] -= millis
            particle[self.SCALE] += self.scaleVelocity * millis
            # TODO: filter out closest collision earlier
            collisions = grid.collideSphere(particle[self.X], 
                                         particle[self.Y],
                                         particle[self.SCALE]*self.radius)
            collision = grid.closestCollision(collisions)
            if collision:
                penetration, tangent, nearestPoint, collider = collision
                # Find tangent direction closest to velocity
                dot = tangent[0]*particle[self.DX] \
                    + tangent[1]*particle[self.DY]
                if dot < 0:
                    tangent = -tangent[0], -tangent[1]
                # Normalize against current velocity
                mag = math.sqrt(tangent[0]**2 + tangent[1]**2)
                tangent = tangent[0]/mag, tangent[1]/mag
                mag = math.sqrt(particle[self.DX]**2 + particle[self.DY]**2)
                tangent = tangent[0]*mag, tangent[1]*mag
                particle[self.DX] = particle[self.DX]*0.2 + tangent[0]*0.8
                particle[self.DY] = particle[self.DY]*0.2 + tangent[1]*0.8
                sz = particle[self.SCALE]*self.radius + penetration
                particle[self.SCALE] = sz/self.radius
                # Some chance that a particle hitting a condensable surface
                # will condense.
                if not particle[self.CONDENSED]:
                    response =  collider.respondSteam(collider, nearestPoint)
                    # When a steam particle condenses into a droplet it
                    # doesn't get removed from the world (it looks ugly),
                    # instead it is marked as having condensed, and can't
                    # condense again until it is reincarnated.
                    if response == collide.Behaviour.STEAM_CONDENSE:
                        particle[self.CONDENSED] = True
            if particle[self.TTL] < 0:
                particle[self.ACTIVE] = False

class Tool(collide.PolygonCollider):
    def __init__(self, factoryMethod, button, disabled, active, pos,
                 allowed=True):
        self.buttonSprite = Sprite(button, pos)
        hx, hy = self.buttonSprite.hx, self.buttonSprite.hy
        x, y = pos
        collide.PolygonCollider.__init__(self, [
            (x-hx, y-hy), (x+hx, y-hy), (x+hx, y+hy), (x-hx, y+hy)])

        self.disabledSprite = Sprite(disabled, pos)
        self.activeSprite = Sprite(active, pos)
        self.factoryMethod = factoryMethod
        self.active = False
        self.disabled = not allowed

        self.selectable = False
        self.hoverable = False
        self.shadow = None

    def draw(self):
        if self.active:
            self.activeSprite.draw()

        if self.disabled:
            self.disabledSprite.draw()
        else:
            self.buttonSprite.draw()

    def drawShadow(self):
        """ Shadow is what appears when tool is selected and mouse hovering
        in work area """
        if self.shadow:
            self.disabledSprite.draw(self.shadow[0], self.shadow[1])

    def drawSelection(self):
        pass

    def place(self, x, y):
        self.factoryMethod(x, y)
        game.selectTool(None)
        
    def beginDrag(self, x, y, alternateButton):
        if self.disabled:
            return
        if game.selectedTool == self:
            game.selectTool(None)
        else:
            game.selectTool(self)

    def continueDrag(self, x, y):
        pass

    def completeDrag(self, x, y):
        pass

class Game:
    def __init__(self):
        self.quit = False
        self.windowSize = (800, 600)

    def init(self):
        pygame.display.init()
        pygame.display.set_caption('Steam')
        surface = pygame.display.set_mode(self.windowSize, 
                                          pygame.DOUBLEBUF |\
                                          pygame.OPENGL)
        self.reshape()
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        self.dropGrid = collide.UniformGridCollider(800, 600, 30, 30)
        self.steamGrid = collide.UniformGridCollider(800, 600, 30, 30)
        self.interactiveGrid = collide.UniformGridCollider(800, 600, 10, 10)
        self.staticPolygons = []

        self.frame = Sprite('frame.png')
        self.frame.pos = (400, 300)

        self.tools = [
            Tool(self.placeMetalObject, 
                 'tool_metal.png', 'toold_metal.png', 'tool_active.png',
                 (750,542)),
            Tool(self.placeWoodObject,
                 'tool_wood.png', 'toold_wood.png', 'tool_active.png',
                 (750,456)),
            Tool(None,
                 'tool_hotplate.png', 'toold_hotplate.png', 'tool_active.png',
                 (750,370), False),
            Tool(None,
                 'tool_kettle.png', 'toold_kettle.png', 'tool_active.png',
                 (750,284), False),
            Tool(None,
                 'tool_lines.png', 'toold_lines.png', 'tool_active.png',
                 (750,198), False),
        ]
        for tool in self.tools:
            tool.addToGrid(self.interactiveGrid)
        self.selectedTool = None


        self.metal = MetalBehaviour(self.dropGrid, self.steamGrid)
        self.wood = WoodBehaviour(self.dropGrid, self.steamGrid)
        self.vaporise = VaporiseBehaviour(self.dropGrid, self.steamGrid)
        
        p = LineStrip([
            (200,400), (600,500), (600, 600), (200, 500)], self.metal)
        p.addInteractiveToGrid(self.interactiveGrid)
        self.staticPolygons.append(p)

        p = LineStrip([
            (100, 100), (100, 300), (300, 100)], self.metal)
        p.addInteractiveToGrid(self.interactiveGrid)
        self.staticPolygons.append(p)

        # Vertical block
        p = LineStrip([
            (100,500), (100, 350), (200, 350), (200, 500)], self.metal)
        p.addInteractiveToGrid(self.interactiveGrid)
        self.staticPolygons.append(p)
        
        # Hot plate
        p = LineStrip([
            (350,50), (450, 50), (450, 20), (350, 20)], self.vaporise)
        p.addInteractiveToGrid(self.interactiveGrid)
        self.staticPolygons.append(p)

        for poly in self.staticPolygons:
            poly.updateGrids()
        
        self.particles = Particles('steam2.png', 
                                   count=500, 
                                   rate=20, 
                                   lifetime=8000,
                                   origin=(400, 50), 
                                   velocity=(0, 0.05),
                                   velocitySigma=(0.01, 0.01),
                                   buoyancy=(0,0.05),
                                   buoyancyAcceleration=0.0005,
                                   scaleVelocity=0.0004)

        self.droplets = Droplets('drip3.png',
                                 count=500,
                                 tangentAcceleration=0.00006,
                                 acceleration=(0,-0.0006),
                                 drag=(0.001, 0))

        self.plants = [Plant(300, 300, self.dropGrid)]
        for plant in self.plants:
            plant.addToGrid(self.interactiveGrid)
            plant.updateGrids()
        self.placedObjects = []

        self.selectedObject = None
        self.hoverObject = None

    def placeMetalObject(self, x, y):
        self.selectedObject = MetalObject(x, y, self.metal)
        self.selectedObject.addToGrid(self.interactiveGrid)
        self.placedObjects.append(self.selectedObject)

    def placeWoodObject(self, x, y):
        self.selectedObject = WoodObject(x, y, self.wood)
        self.selectedObject.addToGrid(self.interactiveGrid)
        self.placedObjects.append(self.selectedObject)

    def selectTool(self, tool):
        if self.selectedTool:
            self.selectedTool.active = False
            self.selectedTool.shadow = None
        self.selectedTool = tool
        if self.selectedTool:
            self.selectedTool.active = True

    def reshape(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, self.windowSize[0], 0, self.windowSize[1], -10, 10)
        glMatrixMode(GL_MODELVIEW)

    def update(self, millis):
        # TODO move grid into particles, droplets?
        self.particles.update(millis, self.steamGrid)
        self.droplets.update(millis, self.dropGrid)
        for plant in self.plants:
            plant.update(millis)

    def display(self):
        glClearColor(1, 1, 1, 1)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.particles.draw()
        self.droplets.draw()
        for poly in self.staticPolygons:
            poly.draw()
        for plant in self.plants:
            plant.draw()
        for object in self.placedObjects:
            object.draw()
        if self.selectedTool:
            self.selectedTool.drawShadow()
        if self.selectedObject:
            self.selectedObject.drawSelection()
        elif self.hoverObject:
            self.hoverObject.drawSelection()
        self.frame.draw()
        for tool in self.tools:
            tool.draw()
        pygame.display.flip()

    def keyboard(self, event):
        if event.key == pygame.K_ESCAPE:
            self.quit = True

    def mouse(self, event):
        x, y = event.pos
        y = 600-y
        dragButton = (hasattr(event, 'buttons') \
                      and (event.buttons[0] | \
                           event.buttons[1] | \
                           event.buttons[2])) \
                     or hasattr(event, 'button')
        collider = self.interactiveGrid.getColliderAt(x, y)

        if self.selectedTool:
            self.selectedTool.shadow = x, y

        if event.type == pygame.MOUSEBUTTONDOWN:
            alternateButton = event.button != 1

            if alternateButton and self.selectedTool:
                self.selectTool(None)
                self.selectedObject = None
            elif self.selectedTool and \
                (not collider or not isinstance(collider, Tool)):
                self.selectedTool.place(x, y)
            elif self.hoverObject and self.hoverObject.keepHover(x,y):
                self.selectedObject = self.hoverObject
            elif collider is not self.selectedObject:
                self.selectedObject = collider
            self.hoverObject = None

            if dragButton and self.selectedObject:
                self.selectedObject.beginDrag(x, y, alternateButton)

        elif event.type == pygame.MOUSEMOTION:
            if dragButton and self.selectedObject:
                self.selectedObject.continueDrag(x, y)
            elif not self.selectedTool:
                if collider and collider.hoverable:
                    self.hoverObject = collider
                elif not (self.hoverObject and self.hoverObject.keepHover(x,y)):
                    self.hoverObject = None
        elif event.type == pygame.MOUSEBUTTONUP:
            if self.selectedObject:
                self.selectedObject.completeDrag(x, y)

    def loop(self):
        clock = pygame.time.Clock()
        millis = 0
        while not self.quit:
            pygame.event.pump()
            for event in pygame.event.get():
                if event.type in (pygame.KEYUP, pygame.KEYDOWN):
                    self.keyboard(event)
                elif event.type in (pygame.MOUSEMOTION, 
                                    pygame.MOUSEBUTTONUP,
                                    pygame.MOUSEBUTTONDOWN):
                    self.mouse(event)
            tick = clock.tick()
            millis += tick
            if millis > 1000:
                sys.stdout.write('FPS %05d\r' % clock.get_fps())
                sys.stdout.flush()
                millis = 0
            self.update(tick)
            self.display()

if __name__ == '__main__':
    global game
    game = Game()
    game.init()
    game.loop()
